"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initialize = exports.Module = exports.Container = exports.Context = exports.newModule = exports.runInContext = exports.ModuleContext = exports.isModuleProvider = exports.createProvider = exports.isModuleProviderValue = void 0;
var tslib_1 = require("tslib");
var isModuleProviderValue = function (input) {
    return !!(input && 'Provider' in input && 'value' in input);
};
exports.isModuleProviderValue = isModuleProviderValue;
function createProvider(defaultValue) {
    var Provider = {
        isModuleProvider: true,
        defaultValue: defaultValue,
        provide: function (value) {
            return {
                Provider: Provider,
                value: value,
            };
        },
    };
    return Provider;
}
exports.createProvider = createProvider;
var isModuleProvider = function (input) {
    return !!(input === null || input === void 0 ? void 0 : input.isModuleProvider);
};
exports.isModuleProvider = isModuleProvider;
var ModuleContext = /** @class */ (function () {
    function ModuleContext() {
        this.deps = {
            modules: new Map(),
            providers: new Map(),
        };
    }
    /**
     * add ModuleCtor instance
     * @param Dep
     * @param module
     * @param replace replace current value or throw error
     */
    ModuleContext.prototype.addModule = function (Ctor, module, replace) {
        if (replace === void 0) { replace = false; }
        if (this.deps.modules.has(Ctor)) {
            var current = this.deps.modules.get(Ctor);
            if (replace) {
                this.deps.modules.set(Ctor, module);
            }
            else if (module !== current) {
                throw new Error("Unexpected duplicate ModuleConstructor: ".concat(Ctor));
            }
            return this;
        }
        this.deps.modules.set(Ctor, module);
        return this;
    };
    /**
     * add Provider value
     * @param Provider
     * @param value
     * @param replace replace current value or throw error
     */
    ModuleContext.prototype.addProvider = function (Provider, value, replace) {
        if (replace === void 0) { replace = false; }
        if (this.deps.providers.has(Provider)) {
            var current = this.deps.providers.get(Provider);
            if (replace) {
                this.deps.providers.set(Provider, value);
            }
            else if (value !== current) {
                throw new Error("Unexpected duplicate Provider");
            }
            return this;
        }
        this.deps.providers.set(Provider, value);
        return this;
    };
    /**
     * get ModuleCtor instance
     * @param ModuleCtor
     */
    ModuleContext.prototype.useModule = function (Ctor) {
        if (this.deps.modules.has(Ctor)) {
            return this.deps.modules.get(Ctor);
        }
        if (Ctor.length > 0) {
            throw new Error("The ModuleConstructor was not found in context: ".concat(Ctor));
        }
        return (0, exports.newModule)(Ctor, this);
    };
    /**
     * get Provider value
     * @param Provider
     */
    ModuleContext.prototype.useProvider = function (Provider) {
        if (this.deps.providers.has(Provider)) {
            return this.deps.providers.get(Provider);
        }
        if (Provider.defaultValue !== undefined) {
            return Provider.defaultValue;
        }
        throw new Error("The Provider is used without injecting or no defaultValue found");
    };
    ModuleContext.prototype.use = function (Dep) {
        if ((0, exports.isModuleProvider)(Dep)) {
            return this.useProvider(Dep);
        }
        return this.useModule(Dep);
    };
    /**
     * inject provider-value
     * @param providerValue
     */
    ModuleContext.prototype.inject = function (providerValue) {
        var Provider = providerValue.Provider, value = providerValue.value;
        this.addProvider(Provider, value);
        return value;
    };
    /**
     * inject modules to module-context
     * @param modules
     */
    ModuleContext.prototype.injectModules = function (modules) {
        var e_1, _a;
        try {
            for (var modules_1 = tslib_1.__values(modules), modules_1_1 = modules_1.next(); !modules_1_1.done; modules_1_1 = modules_1.next()) {
                var module_1 = modules_1_1.value;
                if (module_1.constructor === Object || module_1.constructor === Array || module_1.constructor === Function) {
                    throw new Error("Expected module to be an instance of custom Class, instead of ".concat(JSON.stringify(module_1)));
                }
                this.addModule(module_1.constructor, module_1, true);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (modules_1_1 && !modules_1_1.done && (_a = modules_1.return)) _a.call(modules_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return this;
    };
    /**
     * inject provider-values to module-context
     * @param providers
     */
    ModuleContext.prototype.injectProviderValues = function (providers) {
        var e_2, _a;
        try {
            for (var providers_1 = tslib_1.__values(providers), providers_1_1 = providers_1.next(); !providers_1_1.done; providers_1_1 = providers_1.next()) {
                var _b = providers_1_1.value, Provider = _b.Provider, value = _b.value;
                this.addProvider(Provider, value, true);
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (providers_1_1 && !providers_1_1.done && (_a = providers_1.return)) _a.call(providers_1);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return this;
    };
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    ModuleContext.prototype.new = function (Ctor, options) {
        var e_3, _a;
        var ctx = new ModuleContext();
        try {
            // reusing provider-values of current-context
            for (var _b = tslib_1.__values(this.deps.providers.entries()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var _d = tslib_1.__read(_c.value, 2), Provider = _d[0], value = _d[1];
                ctx.addProvider(Provider, value);
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_3) throw e_3.error; }
        }
        // inject modules
        if (options === null || options === void 0 ? void 0 : options.modules) {
            ctx.injectModules(options.modules);
        }
        // inject provider-values
        if (options === null || options === void 0 ? void 0 : options.providers) {
            ctx.injectProviderValues(options.providers);
        }
        return (0, exports.newModule)(Ctor, ctx);
    };
    return ModuleContext;
}());
exports.ModuleContext = ModuleContext;
/**
 * ModuleContextManager for manager ModuleContext
 */
var currentModuleContext;
var runInContext = function (f, ctx) {
    if (ctx === void 0) { ctx = new ModuleContext(); }
    var prevModuleContext = currentModuleContext;
    try {
        currentModuleContext = ctx;
        var module_2 = f();
        if (module_2 && typeof module_2 !== 'object') {
            throw new Error("Expected function return object, but got ".concat(module_2));
        }
        currentModuleContext.injectModules([module_2]);
        exports.Context.set(module_2, currentModuleContext);
        return module_2;
    }
    finally {
        currentModuleContext = prevModuleContext;
    }
};
exports.runInContext = runInContext;
var newModule = function (Ctor, ctx) {
    if (ctx === void 0) { ctx = new ModuleContext(); }
    if (Ctor.length > 0) {
        throw new Error("Expected ModuleConstructor without parameters, but got ".concat(Ctor));
    }
    return (0, exports.runInContext)(function () { return new Ctor(); }, ctx);
};
exports.newModule = newModule;
var ModuleContextManager = /** @class */ (function () {
    function ModuleContextManager() {
        this.contexts = new WeakMap();
    }
    ModuleContextManager.prototype.set = function (object, ctx) {
        this.contexts.set(object, ctx);
    };
    ModuleContextManager.prototype.from = function (object) {
        if (this.contexts.has(object)) {
            return this.contexts.get(object);
        }
        var context = currentModuleContext !== null && currentModuleContext !== void 0 ? currentModuleContext : new ModuleContext();
        this.set(object, context);
        context.injectModules([object]);
        return context;
    };
    ModuleContextManager.prototype.use = function (Dep) {
        if (!currentModuleContext) {
            throw new Error("Expected use(...) call with runInContext(f)");
        }
        return currentModuleContext.use(Dep);
    };
    /**
     * inject provider-value
     * @param providerValue
     */
    ModuleContextManager.prototype.inject = function (providerValue) {
        if (!currentModuleContext) {
            throw new Error("Expected inject(...) call with runInContext(f)");
        }
        return currentModuleContext.inject(providerValue);
    };
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    ModuleContextManager.prototype.new = function (Ctor, options) {
        if (!currentModuleContext) {
            throw new Error("Expected new(...) call with runInContext(f)");
        }
        return currentModuleContext.new(Ctor, options);
    };
    return ModuleContextManager;
}());
exports.Context = new ModuleContextManager();
/**
 * Container and Module
 */
var Container = /** @class */ (function () {
    function Container() {
        /**
         * add instance to current-context with Constructor as key
         */
        exports.Context.from(this).addModule(this.constructor, this);
    }
    Container.from = function (Ctor) {
        return /** @class */ (function (_super) {
            tslib_1.__extends(InjectableClass, _super);
            function InjectableClass() {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                var _this = _super.apply(this, tslib_1.__spreadArray([], tslib_1.__read(args), false)) || this;
                /**
                 * add instance to current-context with Constructor as key
                 */
                exports.Context.from(_this).addModule(_this.constructor, _this);
                return _this;
            }
            InjectableClass.prototype.use = function (Dep) {
                return exports.Context.from(this).use(Dep);
            };
            /**
             * inject provider-value
             * @param providerValue
             */
            InjectableClass.prototype.inject = function (providerValue) {
                return exports.Context.from(this).inject(providerValue);
            };
            /**
             * create a new context for Dep
             * @param Ctor
             * @param options options for reusing deps or others
             */
            InjectableClass.prototype.new = function (Ctor, options) {
                return exports.Context.from(this).new(Ctor, options);
            };
            return InjectableClass;
        }(Ctor));
    };
    Container.prototype.use = function (Dep) {
        return exports.Context.from(this).use(Dep);
    };
    /**
     * inject provider-value
     * @param providerValue
     */
    Container.prototype.inject = function (providerValue) {
        return exports.Context.from(this).inject(providerValue);
    };
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    Container.prototype.new = function (Ctor, options) {
        return exports.Context.from(this).new(Ctor, options);
    };
    return Container;
}());
exports.Container = Container;
var Module = /** @class */ (function (_super) {
    tslib_1.__extends(Module, _super);
    function Module() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    return Module;
}(Container));
exports.Module = Module;
var initialize = function (Dep, options, ctx) {
    if (ctx === void 0) { ctx = new ModuleContext(); }
    return ctx.new(Dep, options);
};
exports.initialize = initialize;
//# sourceMappingURL=module.js.map