/**
 * ModuleProvider for inject interface/data
 */
export type ModuleProvider<T = any> = {
    isModuleProvider: true;
    defaultValue?: T;
    provide(value: T): ModuleProviderValue<T>;
};
export type ModuleProviderValue<T = any> = {
    Provider: ModuleProvider<T>;
    value: T;
};
export declare const isModuleProviderValue: <T = any>(input: any) => input is ModuleProviderValue<T>;
export declare function createProvider<T>(defaultValue?: T): ModuleProvider<T>;
export declare const isModuleProvider: <T>(input: any) => input is ModuleProvider<T>;
/**
 * ModuleContext for manager dependencies
 */
export type ModuleCtor<T = any> = new (...args: any[]) => T;
export type Injectable<T = any> = ModuleCtor<T> | ModuleProvider<T>;
export type ModuleContextOptions = {
    modules?: object[];
    providers?: ModuleProviderValue[];
};
export declare class ModuleContext {
    deps: {
        modules: Map<ModuleCtor<any>, unknown>;
        providers: Map<ModuleProvider<any>, unknown>;
    };
    /**
     * add ModuleCtor instance
     * @param Dep
     * @param module
     * @param replace replace current value or throw error
     */
    addModule<T>(Ctor: ModuleCtor<T>, module: T, replace?: boolean): this;
    /**
     * add Provider value
     * @param Provider
     * @param value
     * @param replace replace current value or throw error
     */
    addProvider<T>(Provider: ModuleProvider<T>, value: T, replace?: boolean): this;
    /**
     * get ModuleCtor instance
     * @param ModuleCtor
     */
    useModule<T>(Ctor: ModuleCtor<T>): T;
    /**
     * get Provider value
     * @param Provider
     */
    useProvider<T>(Provider: ModuleProvider<T>): T;
    /**
     * get dep by Dep key
     * @param Dep
     */
    use<T>(Dep: ModuleCtor<T>): T;
    use<T>(Dep: ModuleProvider<T>): T;
    /**
     * inject provider-value
     * @param providerValue
     */
    inject<T>(providerValue: ModuleProviderValue<T>): T;
    /**
     * inject modules to module-context
     * @param modules
     */
    injectModules(modules: object[]): this;
    /**
     * inject provider-values to module-context
     * @param providers
     */
    injectProviderValues(providers: ModuleProviderValue[]): this;
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    new<T>(Ctor: ModuleCtor<T>, options?: ModuleContextOptions): T;
}
export declare const runInContext: <T>(f: () => T, ctx?: ModuleContext) => T;
export declare const newModule: <T>(Ctor: ModuleCtor<T>, ctx?: ModuleContext) => T;
declare class ModuleContextManager {
    contexts: WeakMap<object, ModuleContext>;
    set(object: object, ctx: ModuleContext): void;
    from(object: object): ModuleContext;
    /**
     * get dep by Dep key
     * @param Dep
     */
    use<T>(Dep: ModuleCtor<T>): T;
    use<T>(Dep: ModuleProvider<T>): T;
    /**
     * inject provider-value
     * @param providerValue
     */
    inject<T>(providerValue: ModuleProviderValue<T>): T;
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    new<T>(Ctor: ModuleCtor<T>, options?: ModuleContextOptions): T;
}
export declare const Context: ModuleContextManager;
/**
 * Container and Module
 */
export declare class Container {
    static from<T extends ModuleCtor>(Ctor: T): {
        new (...args: any[]): {
            [x: string]: any;
            /**
             * get dep by Dep key
             * @param Dep
             */
            use<T>(Dep: ModuleCtor<T>): T;
            /**
             * get dep by Dep key
             * @param Dep
             */
            use<T_1>(Dep: ModuleProvider<T_1>): T_1;
            /**
             * inject provider-value
             * @param providerValue
             */
            inject<T_2>(providerValue: ModuleProviderValue<T_2>): T_2;
            /**
             * create a new context for Dep
             * @param Ctor
             * @param options options for reusing deps or others
             */
            new<T_3>(Ctor: ModuleCtor<T_3>, options?: ModuleContextOptions): T_3;
        };
    } & T;
    constructor();
    /**
     * get dep by Dep key
     * @param Dep
     */
    use<T>(Dep: ModuleCtor<T>): T;
    use<T>(Dep: ModuleProvider<T>): T;
    /**
     * inject provider-value
     * @param providerValue
     */
    inject<T>(providerValue: ModuleProviderValue<T>): T;
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    new<T>(Ctor: ModuleCtor<T>, options?: ModuleContextOptions): T;
}
export declare class Module extends Container {
}
export declare const initialize: <T>(Dep: ModuleCtor<T>, options?: ModuleContextOptions, ctx?: ModuleContext) => T;
export {};
