export const isModuleProviderValue = (input) => {
    return !!(input && 'Provider' in input && 'value' in input);
};
export function createProvider(defaultValue) {
    const Provider = {
        isModuleProvider: true,
        defaultValue,
        provide(value) {
            return {
                Provider,
                value,
            };
        },
    };
    return Provider;
}
export const isModuleProvider = (input) => {
    return !!(input === null || input === void 0 ? void 0 : input.isModuleProvider);
};
export class ModuleContext {
    constructor() {
        this.deps = {
            modules: new Map(),
            providers: new Map(),
        };
    }
    /**
     * add ModuleCtor instance
     * @param Dep
     * @param module
     * @param replace replace current value or throw error
     */
    addModule(Ctor, module, replace = false) {
        if (this.deps.modules.has(Ctor)) {
            const current = this.deps.modules.get(Ctor);
            if (replace) {
                this.deps.modules.set(Ctor, module);
            }
            else if (module !== current) {
                throw new Error(`Unexpected duplicate ModuleConstructor: ${Ctor}`);
            }
            return this;
        }
        this.deps.modules.set(Ctor, module);
        return this;
    }
    /**
     * add Provider value
     * @param Provider
     * @param value
     * @param replace replace current value or throw error
     */
    addProvider(Provider, value, replace = false) {
        if (this.deps.providers.has(Provider)) {
            const current = this.deps.providers.get(Provider);
            if (replace) {
                this.deps.providers.set(Provider, value);
            }
            else if (value !== current) {
                throw new Error(`Unexpected duplicate Provider`);
            }
            return this;
        }
        this.deps.providers.set(Provider, value);
        return this;
    }
    /**
     * get ModuleCtor instance
     * @param ModuleCtor
     */
    useModule(Ctor) {
        if (this.deps.modules.has(Ctor)) {
            return this.deps.modules.get(Ctor);
        }
        if (Ctor.length > 0) {
            throw new Error(`The ModuleConstructor was not found in context: ${Ctor}`);
        }
        return newModule(Ctor, this);
    }
    /**
     * get Provider value
     * @param Provider
     */
    useProvider(Provider) {
        if (this.deps.providers.has(Provider)) {
            return this.deps.providers.get(Provider);
        }
        if (Provider.defaultValue !== undefined) {
            return Provider.defaultValue;
        }
        throw new Error(`The Provider is used without injecting or no defaultValue found`);
    }
    use(Dep) {
        if (isModuleProvider(Dep)) {
            return this.useProvider(Dep);
        }
        return this.useModule(Dep);
    }
    /**
     * inject provider-value
     * @param providerValue
     */
    inject(providerValue) {
        const { Provider, value } = providerValue;
        this.addProvider(Provider, value);
        return value;
    }
    /**
     * inject modules to module-context
     * @param modules
     */
    injectModules(modules) {
        for (const module of modules) {
            if (module.constructor === Object || module.constructor === Array || module.constructor === Function) {
                throw new Error(`Expected module to be an instance of custom Class, instead of ${JSON.stringify(module)}`);
            }
            this.addModule(module.constructor, module, true);
        }
        return this;
    }
    /**
     * inject provider-values to module-context
     * @param providers
     */
    injectProviderValues(providers) {
        for (const { Provider, value } of providers) {
            this.addProvider(Provider, value, true);
        }
        return this;
    }
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    new(Ctor, options) {
        const ctx = new ModuleContext();
        // reusing provider-values of current-context
        for (const [Provider, value] of this.deps.providers.entries()) {
            ctx.addProvider(Provider, value);
        }
        // inject modules
        if (options === null || options === void 0 ? void 0 : options.modules) {
            ctx.injectModules(options.modules);
        }
        // inject provider-values
        if (options === null || options === void 0 ? void 0 : options.providers) {
            ctx.injectProviderValues(options.providers);
        }
        return newModule(Ctor, ctx);
    }
}
/**
 * ModuleContextManager for manager ModuleContext
 */
let currentModuleContext;
export const runInContext = (f, ctx = new ModuleContext()) => {
    const prevModuleContext = currentModuleContext;
    try {
        currentModuleContext = ctx;
        const module = f();
        if (module && typeof module !== 'object') {
            throw new Error(`Expected function return object, but got ${module}`);
        }
        currentModuleContext.injectModules([module]);
        Context.set(module, currentModuleContext);
        return module;
    }
    finally {
        currentModuleContext = prevModuleContext;
    }
};
export const newModule = (Ctor, ctx = new ModuleContext()) => {
    if (Ctor.length > 0) {
        throw new Error(`Expected ModuleConstructor without parameters, but got ${Ctor}`);
    }
    return runInContext(() => new Ctor(), ctx);
};
class ModuleContextManager {
    constructor() {
        this.contexts = new WeakMap();
    }
    set(object, ctx) {
        this.contexts.set(object, ctx);
    }
    from(object) {
        if (this.contexts.has(object)) {
            return this.contexts.get(object);
        }
        const context = currentModuleContext !== null && currentModuleContext !== void 0 ? currentModuleContext : new ModuleContext();
        this.set(object, context);
        context.injectModules([object]);
        return context;
    }
    use(Dep) {
        if (!currentModuleContext) {
            throw new Error(`Expected use(...) call with runInContext(f)`);
        }
        return currentModuleContext.use(Dep);
    }
    /**
     * inject provider-value
     * @param providerValue
     */
    inject(providerValue) {
        if (!currentModuleContext) {
            throw new Error(`Expected inject(...) call with runInContext(f)`);
        }
        return currentModuleContext.inject(providerValue);
    }
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    new(Ctor, options) {
        if (!currentModuleContext) {
            throw new Error(`Expected new(...) call with runInContext(f)`);
        }
        return currentModuleContext.new(Ctor, options);
    }
}
export const Context = new ModuleContextManager();
/**
 * Container and Module
 */
export class Container {
    static from(Ctor) {
        return class InjectableClass extends Ctor {
            constructor(...args) {
                super(...args);
                /**
                 * add instance to current-context with Constructor as key
                 */
                Context.from(this).addModule(this.constructor, this);
            }
            use(Dep) {
                return Context.from(this).use(Dep);
            }
            /**
             * inject provider-value
             * @param providerValue
             */
            inject(providerValue) {
                return Context.from(this).inject(providerValue);
            }
            /**
             * create a new context for Dep
             * @param Ctor
             * @param options options for reusing deps or others
             */
            new(Ctor, options) {
                return Context.from(this).new(Ctor, options);
            }
        };
    }
    constructor() {
        /**
         * add instance to current-context with Constructor as key
         */
        Context.from(this).addModule(this.constructor, this);
    }
    use(Dep) {
        return Context.from(this).use(Dep);
    }
    /**
     * inject provider-value
     * @param providerValue
     */
    inject(providerValue) {
        return Context.from(this).inject(providerValue);
    }
    /**
     * create a new context for Dep
     * @param Ctor
     * @param options options for reusing deps or others
     */
    new(Ctor, options) {
        return Context.from(this).new(Ctor, options);
    }
}
export class Module extends Container {
}
export const initialize = (Dep, options, ctx = new ModuleContext()) => {
    return ctx.new(Dep, options);
};
//# sourceMappingURL=module.js.map